/*
  *********    ArUniMer - firmware    ************
  Universal Measuring Instrument based on Arduino Nano
  Published in a magazine PRAKTICKA ELEKTRONIKA - Amaterske RADIO - www.aradio.cz

  MIT License

  Copyright (c) 2021 Radim and Stanislav Pechal

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
   
  Version history (see first code line below):
  1.0 - First release
*/
char Version[4]="1.0";  // 3 bytes for program version for initial message

// include the library code:
#include <LiquidCrystal.h>
#include <avr/pgmspace.h>
#include <EEPROM.h>

// LCD and keyboard part
// Initialize the library by associating needed LCD interface pin
// with the arduino pin number connected to
const int rs = 4, en = 3, d4 = 6, d5 = 7, d6 = 8, d7 = 9;
LiquidCrystal lcd(rs, en, d4, d5, d6, d7);
// 4 special chars for battery status
byte BattFull[8] = {B01110, B11111, B11111, B11111, B11111, B11111, B11111, B11111};
byte Batt23[8] = {B01110, B10001, B10001, B11111, B11111, B11111, B11111, B11111};
byte Batt13[8] = {B01110, B10001, B10001, B10001, B10001, B11111, B11111, B11111};
byte BattEmpty[8] = {B01110, B10001, B10001, B10001, B10001, B10001, B10001, B11111};
// and special char for running record
byte RecON[8] = {B010000, B11000, B11100, B11110, B11100, B11000, B10000, B00000};
// Button pin numbers:
const int buttSTST = 12, buttDOWN = 11, buttUP = 10, buttOK = 2;

// Global variables
unsigned char ProbeNum; // Connected probe number
unsigned char MeasQuant = 1;  // Number of measured quantities
char Unit1[4];  // 3 bytes for units of the first measured value
char Unit2[4]="xxx";  // 3 bytes for units of the second measured value
float MeasVal1; // Measured value 1 to display and save
float MeasVal2; // Measured value 2 to display and save
int DecPlac1 = 1;  // Number of decimal places for Value 1 - default is 1 decimal place
int DecPlac2 = 1;  // Number of decimal places for Value 2
int ZeroVal = 0;  // A few probes need zero value
unsigned char datalogRun = 0; // Flag for data logging

// *****************************************************************************
//  Part with setup and measure procedures for probes
// *****************************************************************************
// 1) Here include needed common library code:
// Library for OneWire bus
#include <OneWire.h>
// Library for I2C
#include <Wire.h>

// General purpose function - measure analog value on analog input with 50 Hz filter 
int AverageAi(int anInp) {
  int auxVal = 0;
  for(int i=0; i<18; i++) { // 18 measurements will last about 20 milisec
    auxVal += analogRead(anInp); delay(1);
  }
  return(auxVal/18);
}

// 2) Here define setup and measure functions for probe:
// Both must be void with no parameters
// **********************************************************
// Probe number: 0 - No probe connected
// **********************************************************
void SUNoProbe() {} // No initialization

// No probe - no value
void NoProbe() {
  MeasVal1=(float)0;
}

// **********************************************************
// Probe number: 1 - Voltage on A0
// **********************************************************
void SUVoltage() { DecPlac1 = 2; } // Set number of decimal places

// Just measure voltage on A0 and multiply it by constant (0.0048876 * 10 for 1:10)
void Voltage() {
  MeasVal1=(float)0.048505 * analogRead(A0); // Change constant according to real resistors
  delay(200); // Not to be so fast
}

// **********************************************************
// Probe number: 2 - Current sensor on A1
// **********************************************************
void SUCurrent() {  // Get value for Zero Amps
  DecPlac1 = 2;
  lcd.setCursor(0,1);
  lcd.print("Press OK at 0 A");
  while (digitalRead(buttOK)==HIGH);
  ZeroVal = AverageAi(A1); 
}

// Just measure voltage on A1, subtract ZeroVal and multiply/divide it by constant
void Current() {
  MeasVal1 = (AverageAi(A1) - ZeroVal)/(float)39.44; // Change constant for appropriate sensor range
  delay(200); // Not to be so fast
}

// **********************************************************
// Probe number: 3 - Temperature sensor DS18B20 on D3
// **********************************************************
#include <DallasTemperature.h>  // OneWire.h already included above
// Instance oneWireDS from OneWire library on pin D5
OneWire oneWireDS(5);
// Instance sensorDS from DallasTemperature library
DallasTemperature sensorDS(&oneWireDS);
void SUDS18B20() {
  // Start communication with DS18B20
  sensorDS.begin();
}

// Measure temperature from DS18B20
void DS18B20sen() {
  // Get information from all sensors
  sensorDS.requestTemperatures();
  // And save it
  MeasVal1 = sensorDS.getTempCByIndex(0);
}

// **********************************************************
// Probe number: 4 - Lux sensor BH1750 on I2C
// **********************************************************
// Library for BH1750 sensor (for I2C already above)
#include <BH1750FVI.h>
// Pointer to instance BH1750 sensor from library
BH1750FVI* luxSensor;

void SUBH1750() {
  // Start communication with BH1750
  Wire.begin();
  luxSensor = new BH1750FVI(0x23);
  luxSensor->powerOn();
  luxSensor->setContLowRes();
  DecPlac1 = 0;
}

// Measure intensity from BH1750
void BH1750sen() {
  // Get information from sensor
  MeasVal1 =  luxSensor->getLux();
  delay(200); // Not to be so fast
}

// **********************************************************
// Probe number: 5 - Humidity/temperature sensor AM2320 on I2C
// **********************************************************
// Library AM2320 humidity / temperature sensor
#include <AM2320.h>
// Pointer to instance AM2320 sensor from library
AM2320* HTsens;
void SUAM2320() { MeasQuant = 2; Unit2[0] = '%'; Unit2[1] = 'R'; Unit2[2] = 'H';  // Units for second quantity
   HTsens = new AM2320();
}

// Read humidity and temperature from sensor (only if no error) 
void AM2320sen() {
  if(HTsens->Read()) { MeasVal1=(float)0; MeasVal2=(float)0; } // Output for error code
  else { MeasVal1=HTsens->t; MeasVal2=HTsens->h; }
}

// **********************************************************
// Probe number: 6 - Magnetic induction
// **********************************************************
void SUSS49E() {  // Get value for Zero Tesla
  lcd.setCursor(0,1);
  lcd.print("Press OK at 0 T");
  while (digitalRead(buttOK)==HIGH);
  ZeroVal =  AverageAi(A0); 
}

// Just measure voltage on A0 and multiply it by constant
void SS49E() {
  MeasVal1 = ( AverageAi(A0) - ZeroVal ) *(float)0.3488; // Change constant for appropriate sensor range
  delay(200); // Not to be so fast
}

// **********************************************************
// Probe number: 7 - Barometric pressure BPM180
// **********************************************************
// Library Adafruit for BMP180
#include <Adafruit_BMP085.h>
// Pointer to instance BMP180 sensor from library BMP085
Adafruit_BMP085* BMP180;
void SUBMP180() {  // Start communication with BMP180
  BMP180 = new Adafruit_BMP085();
  BMP180->begin(); 
}

// Read value and divide it by constant to get hPa
void BMP180s() {
  // You can add constant pressure for shift to sea level
  MeasVal1 = BMP180->readPressure()/100.0; 
}

// **********************************************************
// Probe number: 8 - Thermocouple with MAX6675
// **********************************************************
#include <max6675.h>
// Instance for thermocouple sensor from library MAX6675
MAX6675* Thermocouple;
void SUThermcp() {  
  Thermocouple = new MAX6675(A5,A4,5);
  DecPlac1 = 0; // Set decimal places number to zero
}

// Get Celsius temperature from MAX6675
void Thermcp() {
  MeasVal1 = Thermocouple->readCelsius(); // Read temperature
}

// **********************************************************
// Probe number: 9 - 50 kPa pressure sensor on A0
// **********************************************************
void SUPres50() {  // Get value for Zero pressure
  DecPlac1 = 2;
  lcd.setCursor(0,1);
  lcd.print("Press OK at 0 Pa");
  while (digitalRead(buttOK)==HIGH);
  ZeroVal = AverageAi(A0); 
}

// Measure voltage on A0 and multiply it by constant
void Press50() {
  // Change constant for appropriate sensor range
  MeasVal1 = (AverageAi(A0) - ZeroVal) * (float)0.0598; 
  delay(200); // Not to be so fast
}

// **********************************************************
// Probe number: 10 - Battery tester with 50 mA and 0.5A load 
// **********************************************************
void SUBatTes() {  // Prepare parameters
  DecPlac1 = 2; DecPlac2 = 2; MeasVal2 = 0.0;
  pinMode(5, OUTPUT); digitalWrite(5, LOW);   // And stop load
  lcd.setCursor(0,1);
  lcd.print("  OK for start");
  delay(2000);
  MeasQuant = 2; Unit2[0] = ' '; Unit2[1] = 'A'; Unit2[2] = 'h';
}

unsigned char BatSymb=0; 
bool runtest=false;   // Flag for running test 
float StartVolt=1.0; // Start voltage
unsigned long StartTime;  // Start time
void BatTest() {
  // Measure and display battery voltage - always
  MeasVal1 = (float)0.02248 * analogRead(A0); // Change constant for appropriate input divider
  // Test voltage drop on battery
  if(MeasVal1 < (StartVolt*0.8)) runtest=false;
  // If battery test is not running then it can be started
  if (!runtest) {
    digitalWrite(5, LOW);
    if (digitalRead(buttOK)==LOW) {
      if(analogRead(A1) > 512) ZeroVal = 1; else ZeroVal = 0; // Use ZeroVal as flag for load 0.5A or 0.05A
      // Start test, save start time and voltage
      digitalWrite(5, HIGH);
      delay(100);
      StartVolt = (float)0.02248 * analogRead(A0);
      StartTime = millis(); runtest=true;
      // If no battery - then stop it!
      if(StartVolt<1.0) {
        lcd.setCursor(0,1);
        lcd.print("   No battery!  ");
        delay(2000); runtest=false;
      }
    }
  } else
  // If battery test is running then display battery capacity
  {
    int AuxVal; if(analogRead(A1) > 512) AuxVal = 1; else AuxVal = 0;
    if(ZeroVal != AuxVal) runtest=false; // Load was changed - stop test!
    if(ZeroVal == 0) MeasVal2 = (millis() - StartTime)/72000000.0; 
      else MeasVal2 = (millis() - StartTime)/7200000.0;
  }
  delay(200); // Not to be so fast
}

// **********************************************************
// Probe number: 11 - Sound level meter on A0
// **********************************************************
void SUSndLev() {} // No setup is needed

// Just measure voltage on A0 shift and multiply it by constant
void SoundLev() {
  int auxVal = 0;
  for(int i=1; i<10; i++){ auxVal += analogRead(A0); delay(20); }
  MeasVal1=(float)auxVal / 41.93 + 25.36;   // Adjust the constants according to the reference instrument
}

// **********************************************************
// Common part for numbers: 12,13,14 - Frequency meters
// **********************************************************
// Timer 1 overflows counter
volatile unsigned char overcount;
ISR(TIMER1_OVF_vect) {
  overcount++; // increment overflow counter
}
void SUFreq(){ // Common setup for frequency meters
  if(ProbeNum<14) DecPlac1 = 2;
  overcount = 0; // reset overflow counter
  // Set control registers (see datasheet)
  TCCR1A = 0; // normal mode of operation
  TIMSK1 = bit(TOIE1); // enable interrupt on overflow
}

// **********************************************************
// Probe number: 12 - Frequency meter up to 100 MHz
// **********************************************************
void FreqMHZ() {
  TCNT1 = 0; overcount = 0; // Clear timer
  TCCR1B = bit(CS12) | bit(CS11) | bit(CS10); // Start timer with external clock source
  delay(100);
  TCCR1B = 0; // Stop timer
  unsigned long totalSamples = overcount;
  totalSamples = (totalSamples << 16) + TCNT1;  
  MeasVal1 = totalSamples * 0.00032; // Sampling period is 100 msec - return value in MHz
  delay(200); // For better read
}

// **********************************************************
// Probe number: 13 - Frequency meter up to 5 MHz - display kHz
// **********************************************************
void FreqkHZ() {
  TCNT1 = 0; overcount = 0; // Clear timer
  TCCR1B = bit(CS12) | bit(CS11) | bit(CS10); // Start timer with external clock source
  delay(400);
  TCCR1B = 0; // Stop timer
  unsigned long totalSamples = overcount;
  totalSamples = (totalSamples << 16) + TCNT1;  
  MeasVal1 = totalSamples * 0.0025; // Sampling period is 400 msec - return value in kHz
}

// **********************************************************
// Probe number: 14 - Frequency meter up to 5 MHz - display Hz
// **********************************************************
void FreqHZ() {
  TCNT1 = 0; overcount = 0; // Clear timer
  TCCR1B = bit(CS12) | bit(CS11) | bit(CS10); // Start timer with external clock source
  delay(400);
  TCCR1B = 0; // Stop timer
  unsigned long totalSamples = overcount;
  totalSamples = (totalSamples << 16) + TCNT1;  
  MeasVal1 = totalSamples * 2.5; // Sampling period is 400 msec - return value in Hz
  if((totalSamples>2)&(totalSamples<100)) { // For low freq try measure period instead
    unsigned long ontime = pulseIn(5,HIGH);
    unsigned long offtime = pulseIn(5,LOW);
    MeasVal1 = 994000.0 / (ontime + offtime); // Frequency in Hz
  }
}

// **********************************************************
// Probe number: 15 - Ultrasonic distance meter with HC-SR04
// **********************************************************
void SUDistan() {
  // Set pins for Trig and Echo
  pinMode(A1, OUTPUT);
  pinMode(5, INPUT);
}

// Measure pulse duration and calculate distance
void Distan() {
  // Start pulse for module HC-SR04
  digitalWrite(A1, LOW);
  delayMicroseconds(10);
  digitalWrite(A1, HIGH);
  delayMicroseconds(10);
  digitalWrite(A1, LOW);
  // Get time for sound reflection
  long duration = pulseIn(5, HIGH);
  // And calculate distance
  MeasVal1=(float)duration*0.0171;
  // Sound speed = 343 m/sec (at 20C) and half it for 2 ways
  delay(200); // For better read
}

// **********************************************************
// Probe number: 16 - Infrared temperature sensor MLX90614 on I2C
// **********************************************************
// Library for MLX90614 sensor
#include <Adafruit_MLX90614.h>
// Pointer to instance MLX90614 sensor from library
Adafruit_MLX90614* infraTemp;

void SUInfrared() {
  // Start communication with MLX90614
  infraTemp = new Adafruit_MLX90614();
  infraTemp->begin();
}

// Just measure object temperature from MLX90614
void Infrared() {
  // Get value from sensor
  MeasVal1=infraTemp->readObjectTempC();
  delay(200); // For better read
}


// ******************************************************
// 3) Here add your functions to tables (SUprobe - setup, Probe - measure)
typedef void (* p_func)();
p_func SUProbe[] = {SUNoProbe,SUVoltage,SUCurrent,SUDS18B20,SUBH1750,SUAM2320,SUSS49E,SUBMP180,SUThermcp,SUPres50,SUBatTes,SUSndLev,SUFreq,SUFreq,SUFreq,SUDistan,SUInfrared,SUNoProbe,SUNoProbe,SUNoProbe,SUNoProbe,SUNoProbe,SUNoProbe,SUNoProbe,SUNoProbe}; 
p_func Probe[] = {NoProbe,Voltage,Current,DS18B20sen,BH1750sen,AM2320sen,SS49E,BMP180s,Thermcp,Press50,BatTest,SoundLev,FreqMHZ,FreqkHZ,FreqHZ,Distan,Infrared,NoProbe,NoProbe,NoProbe,NoProbe,NoProbe,NoProbe,NoProbe,NoProbe};

// ******************************************************
// 4) And finally add units for probe to array table (just for the first quantity)
  const char unit_0[] PROGMEM = " - ";
  const char unit_1[] PROGMEM = " V ";
  const char unit_2[] PROGMEM = " A ";
  const char unit_3[] PROGMEM = "\337C ";   // 337 = octal value for degree on LCD 
  // Note: 337 octal value is only for LCD with Asian charset, use 260 octal value for LCD with European charset 
  const char unit_4[] PROGMEM = " Lx";
  const char unit_5[] PROGMEM = "\337C ";   // 337 = octal value for degree on LCD
  const char unit_6[] PROGMEM = " mT";
  const char unit_7[] PROGMEM = "hPa";
  const char unit_8[] PROGMEM = "\337C ";
  const char unit_9[] PROGMEM = "kPa";
  const char unit_10[] PROGMEM = " V ";
  const char unit_11[] PROGMEM = " dB";
  const char unit_12[] PROGMEM = "MHz";
  const char unit_13[] PROGMEM = "kHz";
  const char unit_14[] PROGMEM = " Hz";
  const char unit_15[] PROGMEM = " cm";
  const char unit_16[] PROGMEM = "\337C ";   // 337 = octal value for degree on LCD
  const char unit_17[] PROGMEM = " - ";
  const char unit_18[] PROGMEM = " - ";
  const char unit_19[] PROGMEM = " - ";
  const char unit_20[] PROGMEM = " - ";
  const char unit_21[] PROGMEM = " - ";
  const char unit_22[] PROGMEM = " - ";
  const char unit_23[] PROGMEM = " - ";
  const char unit_24[] PROGMEM = " - ";
  const char* const unit_table[] PROGMEM = {unit_0, unit_1, unit_2, unit_3, unit_4, unit_5, unit_6, unit_7, unit_8, unit_9, unit_10, unit_11, unit_12, unit_13, unit_14, unit_15, unit_16, unit_17, unit_18, unit_19, unit_20, unit_21, unit_22, unit_23, unit_24};
// *****************************************************************************
// End of probes part 
// *****************************************************************************


// ******************************************************
// Datalog tables and functions
// ******************************************************
  unsigned char datalogDelayID = 0;
  char datalogDelayTextShow[6];
  const char datalogDelayText_0[] PROGMEM = " 5 sec";
  const char datalogDelayText_1[] PROGMEM = "10 sec";
  const char datalogDelayText_2[] PROGMEM = "15 sec";
  const char datalogDelayText_3[] PROGMEM = "30 sec";
  const char datalogDelayText_4[] PROGMEM = " 1 min";
  const char datalogDelayText_5[] PROGMEM = " 2 min";
  const char datalogDelayText_6[] PROGMEM = " 5 min";
  const char datalogDelayText_7[] PROGMEM = "10 min";
  const char datalogDelayText_8[] PROGMEM = "15 min";
  const char datalogDelayText_9[] PROGMEM = "30 min";
  const char datalogDelayText_10[] PROGMEM = "1 hour";
  const char* const datalogDelayText[] PROGMEM = {datalogDelayText_0, datalogDelayText_1, datalogDelayText_2, datalogDelayText_3, datalogDelayText_4, datalogDelayText_5, datalogDelayText_6, datalogDelayText_7, datalogDelayText_8, datalogDelayText_9, datalogDelayText_10};
  uint16_t datalogDelaySec[11] = { 5, 10, 15, 30, 60, 120, 300, 600, 900, 1800, 3600};

  unsigned char datalogMemoryPointer = 0;
  unsigned long datalogLastTime;

void showDatalogDelayTime () {
  lcd.clear();  
  lcd.print("Datalog delay   ");    
  lcd.setCursor(0,1);
  lcd.print("time: ");
  strcpy_P(datalogDelayTextShow, (char*)pgm_read_word(&(datalogDelayText[datalogDelayID])));  
  lcd.print(datalogDelayTextShow);
}
// ******************************************************


// Test and return number of connected probe
unsigned char GetProbeNumber() {
  unsigned char rProbe = 0;
  int ADval = analogRead(A2); // Read and evaluate voltage on the first pin
  if(ADval > 205) rProbe++; if(ADval > 410) rProbe++; if(ADval > 605) rProbe++; if(ADval > 820) rProbe++;
  ADval = analogRead(A3);; // Read and evaluate voltage on the second pin
  if(ADval > 205) rProbe+=(unsigned char)5; if(ADval > 410) rProbe+=(unsigned char)5; if(ADval > 605) rProbe+=(unsigned char)5; if(ADval > 820) rProbe+=(unsigned char)5;
  return(rProbe);  
}

// Device initial setup
void setup() {
  // Set up the LCD's number of columns and rows:
  lcd.createChar(0, BattFull);
  lcd.createChar(1, Batt23);
  lcd.createChar(2, Batt13);
  lcd.createChar(3, BattEmpty);
  lcd.createChar(4, RecON);
  lcd.begin(16, 2);
  // Define button pin mode
  pinMode(buttSTST, INPUT_PULLUP);
  pinMode(buttDOWN, INPUT_PULLUP);
  pinMode(buttUP, INPUT_PULLUP);
  pinMode(buttOK, INPUT_PULLUP);
  // Get type of connected probe - must be the same during measure
  ProbeNum = GetProbeNumber();
  // Setup probe for measurement
  SUProbe[ProbeNum]();
  // And prepare units for probe
  strcpy_P(Unit1, (char*)pgm_read_word(&(unit_table[ProbeNum])));
  // Opens serial port, sets data rate to 9600 bps  
  Serial.begin(9600); 
  // datalogMemoryPointer inicialization
  int pointer=20;   // Data start at address 20
  float value;
  EEPROM.get(pointer, value);
  while (!((isnan(value))||(pointer>1019))) {
    pointer+=4;
    datalogMemoryPointer++;
    EEPROM.get(pointer, value);
  }
  // and initial message with verison
  lcd.setCursor(4, 0);  
  lcd.print("ArUniMer");    
  lcd.setCursor(6, 1);
  lcd.print(Version);
  delay(1000);       
}

// *****************************************************************************
// Main device function loop
// *****************************************************************************
int i;  // Auxiliary variable
void loop() {
  if(ProbeNum != GetProbeNumber()) {  // Test for probe disconnection
    lcd.setCursor(0, 1);
    lcd.write("Error connection");
    while(1);        // Disconnection during working - Error and endless loop
  }
  // Section for serial communication
  if (Serial.available()) {
    switch (Serial.read()) {
      // Ping
      case '?':      
        Serial.print((char)0xB6);
      break;
      // Data transfer in text mode
      case 'M': {  
        Serial.print("# unit: ");
        for (i=2;i<5;i++) {Serial.print((char)EEPROM.read(i));}
        if(EEPROM.read(1)==2) {Serial.print("\t"); for (i=6;i<9;i++) {Serial.print((char)EEPROM.read(i));}}
        Serial.println();
        int memDelay = (EEPROM.read(10)<<8)&0xFF00 | (EEPROM.read(11)&0x00FF);
        Serial.print("# interval: ");
        Serial.print(memDelay);
        Serial.println(" sec");
        int pointer=20;
        float value;
        EEPROM.get(pointer, value);        
        i=0;
        while (!(isnan(value))) {
          Serial.print(i*memDelay);          
          i++;
          Serial.print(" \t");                    
          Serial.print(value);          
          pointer+=4;
          if(EEPROM.read(1)==2) {
            if (pointer<1024) EEPROM.get(pointer, value); else break;
            Serial.print(" \t");                    
            Serial.print(value);          
            pointer+=4;
          }
          Serial.println();          
          if (pointer<1024) EEPROM.get(pointer, value); else break;
        }
      }
      break;
      // Data transfer in automatic program download to PC
      case 0xB5:      
        for (i=0;i<1024;i++) {
          Serial.write((unsigned char)EEPROM.read(i));
        }
      break;
    }
  }

  // Section for button manipulation
  // Button UP and DOWN
  if (((digitalRead(buttUP)==LOW) or ((digitalRead(buttDOWN)==LOW))) && (datalogRun==0)) {
    showDatalogDelayTime();
    delay(100);   // wait for 100 milisecond    
    while (digitalRead(buttUP)==LOW); // wait for button release
    while (digitalRead(buttDOWN)==LOW); // wait for button release       
    // If button UP or down not pressed in 5 sec will be continued
    i=0;
    while (i<50) { 
      i++;
      delay(100);   // wait for 100 milisecond    
      // Button UP pressed, datalog time will be increase 
      if (digitalRead(buttUP)==LOW) {
        i=0;
        if (datalogDelayID<10) datalogDelayID++;
        showDatalogDelayTime();
        delay(100);   // wait for 100 milisecond    
        while (digitalRead(buttUP)==LOW); // wait for button release        
      } 
      // Button DONW pressed, datalog time will be decrease 
      else if (digitalRead(buttDOWN)==LOW) {
        i=0;
        if (datalogDelayID>0) datalogDelayID--;        
        showDatalogDelayTime();
        delay(100);   // wait for 100 milisecond    
        while (digitalRead(buttDOWN)==LOW); // wait for button release        
      }
    }    
  }
  // Button Run / Stop
  if (digitalRead(buttSTST)==LOW) {    
    if (datalogRun) {
      datalogRun = 0;
      lcd.setCursor(0, 0);  
      lcd.print("Datalog stopped ");    
      lcd.setCursor(0, 1);
      lcd.print("                ");         
    } else {
      datalogRun = 1;
      lcd.setCursor(0, 0);  
      lcd.print("Datalog started ");    
      lcd.setCursor(0, 1);
      lcd.print("                ");         
      datalogLastTime = millis();
      // Table head
      EEPROM.write(0,ProbeNum); EEPROM.write(1,MeasQuant); EEPROM.write(2,Unit1[0]);      
      EEPROM.write(3,Unit1[1]); EEPROM.write(4,Unit1[2]); EEPROM.write(5,DecPlac1);
      EEPROM.write(6,Unit2[0]); EEPROM.write(7,Unit2[1]); EEPROM.write(8,Unit2[2]);      
      EEPROM.write(9,DecPlac2);
      EEPROM.write(10,(datalogDelaySec[datalogDelayID]&0xFF00)>>8);
      EEPROM.write(11,datalogDelaySec[datalogDelayID]&0x00FF);
      for (i=12;i<20;i++) EEPROM.write(i,'x');      
      // First data
      if(MeasQuant == 2) {
        datalogMemoryPointer = 2;
        EEPROM.put(20, MeasVal1);
        EEPROM.put(24, MeasVal2);
        for (i=28; i<32; i++) EEPROM.write(i,0xFF);
      } else {
        datalogMemoryPointer = 1;
        EEPROM.put(20, MeasVal1);
        for (i=24; i<28; i++) EEPROM.write(i,0xFF);
      }
    }
    delay(1000);   // wait for 1 sec    
    while (digitalRead(buttSTST)==LOW); // wait for button release   
  }

  // Section for printing values to the LCD.
  // Start probe measuring procedure
  Probe[ProbeNum]();
  // Clear LCD and set the cursor to column 0, line 0
  lcd.clear();
  // Display battery voltage symbol
  lcd.setCursor(0,1);
  i = analogRead(A6);
  if (i>550) {lcd.write((unsigned char)0);}       // 8.9V
  else if (i>523) {lcd.write((unsigned char)1);}  // 8.5V
  else if (i>490) {lcd.write((unsigned char)2);}  // 8.0V
  else {lcd.write((unsigned char)3); datalogRun = 0;} // very low voltage stops recording at the same time
// Display probe number (for debug mode only)
//  lcd.write("  ");
//  lcd.setCursor(2,1);
//  lcd.print(ProbeNum);
  // Special part for battery test
  if(runtest) {
    lcd.setCursor(4,1);
    lcd.write(BatSymb++);
    BatSymb=BatSymb & (unsigned char)0x3;
    }
  // Memory percentage usage
  lcd.setCursor(5,1);  
  if (datalogRun) lcd.write(4); else lcd.write(' ');  // Data record running?
  lcd.print("M:");  
  lcd.setCursor(9,1);  
  float memPercentage = (float)(datalogMemoryPointer) / 2.50;
  lcd.print(memPercentage,1);    
  lcd.write('%');
  // For double quantities display both MeasVal on upper line
  if(MeasQuant == 2) {
    lcd.setCursor(0,0);
    lcd.print(MeasVal1, DecPlac1);
    lcd.print(Unit1);
    lcd.setCursor(8,0);
    lcd.print(MeasVal2, DecPlac2);
    lcd.print(Unit2);
  } else {
  // For any other version display only MeasVal1 in the centre of upper line
    if(MeasVal1<0.0) lcd.setCursor(4,0); else lcd.setCursor(5,0);
    // No decimal places for big numbers
    if(MeasVal1>99999.0)lcd.print(MeasVal1, 0); else lcd.print(MeasVal1, DecPlac1);
    lcd.print(Unit1);
  }
  // wait for 300 milisecond and datalog procedure
  long int startmilis = millis();
  while ((millis()-startmilis) < 300) {
    // Datalog
    long int diff = (millis()-datalogLastTime);
    long int datalogDelaymiliSec = datalogDelaySec[datalogDelayID];
    datalogDelaymiliSec *= 1000;
    if ((diff > datalogDelaymiliSec) && (datalogRun)) {
      datalogLastTime+=datalogDelaymiliSec;
      if (datalogMemoryPointer<(251-MeasQuant)) {
        int pointer = datalogMemoryPointer * 4 + 20;
        EEPROM.put(pointer, MeasVal1);       
        datalogMemoryPointer++;    
        if (MeasQuant == 2) {
          pointer += 4;
          EEPROM.put(pointer, MeasVal2);       
          datalogMemoryPointer++;
        }  
        pointer+=3; // And mark end of saved array
        for (i=0; i<4;i++) {
          pointer++;
          EEPROM.write(pointer, 0xFF);  
        }
      } else datalogRun = 0;
    }
  }
}
