#!/usr/bin/python

import struct
import numpy
import matplotlib
import matplotlib.pyplot as graph
import sys
import glob
import serial
import time
from PyQt5 import QtCore, QtWidgets
from PyQt5.QtWidgets import QApplication, QWidget, QLabel, QProgressDialog
from PyQt5.QtGui import QIcon, QPixmap

# Global variables for program
filnames = "UM" + time.strftime("%y%m%d_%H%M%S")
unidata = []   # Buffer for data received from serial port
Period = 1
UnitTime = 'time [x]'
Unit1 = '   '
Unit2 = '   '
nCount = 0
NoError = True

# Look for available ports and try to get data from device
# Code based on hint from Thomas (stackoverflow.com)
def get_data_ser_port():
    global nCount
    """ Lists serial port names
        :raises EnvironmentError:
            On unsupported or unknown platforms
        :returns:
            A list of the serial ports available on the system
    """
    if sys.platform.startswith('win'):
        ports = ['COM%s' % (i + 1) for i in range(100)]
    elif sys.platform.startswith('linux') or sys.platform.startswith('cygwin'):
        # this excludes your current terminal "/dev/tty"
        ports = glob.glob('/dev/tty[A-Za-z]*')
    elif sys.platform.startswith('darwin'):
        ports = glob.glob('/dev/tty.*')
    else:
        raise EnvironmentError('Unsupported platform')

    a = 0  # Auxiliary couter for progress bar
    # Display dialog for progress in seeking of port
    dlg = QtWidgets.QProgressDialog()
    dlg.setWindowModality(QtCore.Qt.WindowModal)
    dlg.forceShow()
    dlg.setWindowTitle('Please wait!')
    dlg.setLabelText('Looking for the ArUniMer...')
    dlg.setCancelButton(None)
    for port in ports:
        a += 1
        try:
            s = serial.Serial(port)
            s.baudrate = 9600
            s.bytesize = 8
            s.parity = 'N'
            s.timeout = 2
            # Display actual progress in seeking of port
            dlg.setValue(a)
            # Opening of serial port with Arduino can take a long time, so wait
            time.sleep(3)
            s.write(b'?')
            if(s.read(1) == b'\xB6'):
               s.write(b'\xB5')
               while True:
                   sdata = s.read(1)
                   if not sdata:
                       # break out of the loop - no data on serial line
                       break
                   unidata.extend(sdata)
                   nCount = nCount + 1
                   if (nCount>1024):
                       # break out of the loop - too many bytes
                       break
            s.close()
        except (OSError, serial.SerialException):
            pass
    dlg.accept()

# Convert characters from LCD chars to PC chars
def CharConv(inpChar):
    if inpChar == 0xDF:
        return chr(0xB0)
    else:
        return chr(inpChar)
# Process received data and save it to files with picture and table
def ProcessData():
    global NoError
    # numpy functions needs an array instead of a list
    values = bytearray(unidata)
    if((values[1]<1) or (values[1]>2)):     # Check type of received data
        NoError = False
        return
    # Get number of measured values
    nCount = 0
    for i in range(250):
        if ((values[i*4 +20] == 255) and (values[i*4 +21] == 255) and (values[i*4 +22] == 255) and (values[i*4 +23] == 255)):
            break
        nCount = nCount + 1
    if(nCount<2):     # Must be 2 values at least
        NoError = False
        return
    # Get period time
    Period = values[10]*256 + values[11]
    if((Period<2) or (Period>8000)):     # Check period range
        NoError = False
        return
    UnitTime = 'Time [s]'
    if (Period * nCount) > 999:   # Change timebase to minutes for 16 minutes and more
        Period = Period / 60
        UnitTime = 'Time [m]'
    if (Period * nCount) > 999:   # Change timebase to hours for 16 hours and more
        Period = Period / 60
        UnitTime = 'Time [h]'
    # Get unit1 and unit2 name
    Unit1 = CharConv(values[2])+CharConv(values[3])+CharConv(values[4])
    Unit2 = CharConv(values[6])+CharConv(values[7])+CharConv(values[8])
    # Prepare X data for plotting 
    if values[1]==1:   # Prepare long array
        npX = numpy.arange(0, (nCount)*Period, Period)
    else:   # Prepare short array
        npX = numpy.arange(0, (nCount/2)*Period, Period)
    # Y data for single graph and double data
    if values[1]==1:   # Prepare single array
        npY1 = numpy.empty(0)
        for k in range(int(nCount)):
            npY1 = numpy.append(npY1, (struct.unpack('<f', values[(k*4+20):(k*4+24)])))
    if values[1]==2:   # Just for double data prepare 2 arrays
        npY1 = numpy.empty(0)
        for k in range(int(nCount/2)):
            npY1 = numpy.append(npY1, (struct.unpack('<f', values[(k*8+20):(k*8+24)])))
        npY2 = numpy.empty(0)
        for k in range(int(nCount/2)):
            npY2 = numpy.append(npY2, (struct.unpack('<f', values[(k*8+24):(k*8+28)])))
    # Prepare graph
    fig, ax1 = graph.subplots(figsize=(12, 7))
    graph.title('Graph of measured values')
    ax1.grid(True)
    color = 'tab:red'
    ax1.set_xlabel(UnitTime)
    ax1.set_ylabel(Unit1, color=color)
    ax1.plot(npX, npY1, color=color)
    ax1.tick_params(axis='y', labelcolor=color)
    if(values[1]==2):
        ax2 = ax1.twinx()  # instantiate a second axes that shares the same x-axis
        color = 'tab:blue'
        ax2.set_ylabel(Unit2, color=color)  # we already handled the x-label with ax1
        ax2.plot(npX, npY2, color=color)
        ax2.tick_params(axis='y', labelcolor=color)
    fig.tight_layout()  # otherwise the right y-label is slightly clipped
    # And save files - picture and "csv" table
    graph.savefig(filnames+'.png', dpi=72)
    if(values[1]==2):
        numpy.savetxt(filnames+'.csv', numpy.column_stack((npX,npY1,npY2)), fmt='%1.1f \t%1.3f \t%1.3f', header='Saved values')
    else:
        numpy.savetxt(filnames+'.csv', numpy.column_stack((npX,npY1)), fmt='%1.1f \t%1.3f', header='Saved values')

# Show window with graph
class App(QWidget):
    def __init__(self):
        super().__init__()
        self.title = 'ArUniMer - data processing'
        self.left = 20
        self.top = 40
        self.width = 640
        self.height = 480
        self.initUI()
    def initUI(self):
        self.setWindowTitle(self.title)
        self.setGeometry(self.left, self.top, self.width, self.height)
        # Create widget
        label = QLabel(self)
        pixmap = QPixmap(filnames+'.png')
        label.setPixmap(pixmap)
        self.resize(pixmap.width(),pixmap.height())
        self.show()

# Main program loop
if __name__ == '__main__':
    app = QApplication(sys.argv)
    get_data_ser_port()
    if nCount != 1024:
        msg = QtWidgets.QMessageBox()
        msg.setText("No device found or data transfer error!  ")
        msg.setWindowTitle("ArUniMer error!")
        msg.exec_()
    else:
        ProcessData()
        if NoError:
            ex = App()
            sys.exit( app.exec_() )
        else:
            msg = QtWidgets.QMessageBox()
            msg.setText("Received data error!")
            msg.setWindowTitle("ArUniMer error!")
            msg.exec_()


